/*
 *  GuiLoader, a GuiXml loader library
 *  Copyright (c) 2006 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "guiloader_impl.h"
#include "proxies.h"

GUILOADER_API gint guiloader_version(void) {
	return GUILOADER_VERSION;
}

GUILOADER_API void guiloader_enter_check(const gchar * msg) {
	GuiLoaderClass * meta= GUI_LOADER_CLASS(g_type_class_ref(GUI_TYPE_LOADER));
	if(!meta->init)
		meta->init= NewInit();
	else
		++meta->init->refcounter;
	g_type_class_unref(meta);
}

GUILOADER_API void guiloader_leave(void) {
	GuiLoaderClass * meta= GUI_LOADER_CLASS(g_type_class_ref(GUI_TYPE_LOADER));
	--meta->init->refcounter;
	if(meta->init->refcounter == 0) {
		FreeInit(meta->init);
		meta->init= NULL;
	}
	g_type_class_unref(meta);
}

GUILOADER_API GType gui_loader_get_type(void) {
	static GType object_type= 0;
	if(object_type==0) {
		static const GTypeInfo object_info= {
			sizeof(GuiLoaderClass), 
			NULL, 
			NULL,	
			(GClassInitFunc)gui_loader_class_init,
			NULL, 
			NULL, 
			sizeof(GuiLoader), 
			0,
			(GInstanceInitFunc)gui_loader_init,
		};
	  object_type= g_type_register_static(G_TYPE_OBJECT, "GuiLoader", &object_info, 0);
	}
	return object_type;
}

GuiLoader * gui_loader_new(void) {
	return g_object_new(GUI_TYPE_LOADER, NULL);
}

GUILOADER_API void gui_loader_set_preview_mode(GuiLoader * loader, gboolean preview) {
	GuiLoaderPrivate * priv;
	Check(GUI_IS_LOADER(loader));
	priv= GUI_LOADER_GET_PRIVATE(loader);
	priv->preview= preview;
	Check(!priv->preview || !priv->steal);
}

GUILOADER_API void gui_loader_set_steal_objects(GuiLoader * loader, gboolean steal) {
	GuiLoaderPrivate * priv;
	Check(GUI_IS_LOADER(loader));
	priv= GUI_LOADER_GET_PRIVATE(loader);
	priv->steal= steal;
	Check(!priv->preview || !priv->steal);
}

GUILOADER_API void gui_loader_load_from_string(GuiLoader * loader, const gchar * data, gssize size) {
	Check(GUI_IS_LOADER(loader) && data);
	FreeGui(loader);
	LoadModel(loader, data, size);
	LoadGui(loader);
	FreeModel(loader);
}

GUILOADER_API gboolean gui_loader_load_from_file(GuiLoader * loader, const gchar * filename, GError ** error) {
	gchar * data;
	gsize size;
	gboolean result;
	Check(GUI_IS_LOADER(loader));
	result= g_file_get_contents(filename, &data, &size, error);
	if(result) {
		gui_loader_load_from_string(loader, data, size);
		g_free(data);
	}
	return result;
}

GUILOADER_API void gui_loader_clear(GuiLoader * loader) {
	Check(GUI_IS_LOADER(loader));
	FreeGui(loader);
}

GUILOADER_API GQueue * gui_loader_query_connectors(GuiLoader * loader, const gchar * name, gint count) {
	GuiLoaderPrivate * priv;
	GQueue * queue;
	Check(GUI_IS_LOADER(loader) && name);
	priv= GUI_LOADER_GET_PRIVATE(loader);
	queue= g_hash_table_lookup(priv->connectors, name);
	Check(queue && g_queue_get_length(queue)==count);
	return queue;
}

GUILOADER_API void gui_loader_get_connect_info(GuiLoader * loader, const GuiConnector * connector, GuiConnectInfo * info) {
	Check(GUI_IS_LOADER(loader) && connector && info);
	info->object= connector->object;
	info->signame= ShortStrGetStr(connector->emitter->signame);
	info->after= connector->emitter->after;
}

GUILOADER_API void gui_loader_connect_signals(GuiLoader * loader, const gchar * name, GCallback func, gpointer userdata, gint count) {
	GQueue * queue= gui_loader_query_connectors(loader, name, count);
	GList * elem;
	for(elem= g_queue_peek_head_link(queue); elem; elem= g_list_next(elem)) {
		GuiConnectInfo info;
		gulong hid;
		gui_loader_get_connect_info(loader, elem->data, &info);
		hid= g_signal_connect_data(
			info.object, 
			info.signame, 
			func, userdata, NULL, 
			info.after ? G_CONNECT_AFTER : 0
		);
		if(!hid)
			FatalError("Connecting signals failed (see log messages)");
	}
}

GUILOADER_API gpointer gui_loader_bind_object(GuiLoader * loader, const gchar * name) {
	GObject * object;
	Check(GUI_IS_LOADER(loader) && name);
	object= FindObject(loader, name);
	Check(object);
	if(GTK_IS_WINDOW(object))
		g_object_set_data(object, "GuiLoaderBoundWindow", "true");
	return object;
}

void gui_loader_class_init(GuiLoaderClass * meta) {
	GObjectClass * object_class;
	meta->init= NULL;
	meta->parent_class= g_type_class_peek_parent(meta);
	object_class= G_OBJECT_CLASS(meta);
	object_class->finalize= gui_loader_finalize;
	g_type_class_add_private(object_class, sizeof(GuiLoaderPrivate));
}

void gui_loader_init(GuiLoader * loader) {
	InitInstance(loader);
}

void gui_loader_finalize(GObject * object) {
  GuiLoader * loader= GUI_LOADER(object);
  GuiLoaderClass * meta= GUI_LOADER_GET_CLASS(loader);
	DoneInstance(loader);
	(*meta->parent_class->finalize)(object);
}

void CheckFailed(const gchar * expr, const gchar * file, unsigned line) {
	gchar msg[256];
	#ifdef GUILOADER_SANDBOX
		assert(0);
	#endif
	g_snprintf(msg, sizeof(msg), "Check failed: %s, file: %s, line: %d", expr, file, line);
	FatalError(msg);
}

void FatalError(const gchar * message) {
	GtkWidget * dialog;
	g_critical(message);

	dialog= gtk_message_dialog_new(
		NULL,
		GTK_DIALOG_MODAL,
		GTK_MESSAGE_ERROR,
		GTK_BUTTONS_CLOSE,
		"Fatal Error [" G_LOG_DOMAIN "]"
	);

	gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), message);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);

	abort();
}

ShortStr * ShortStrNew(const gchar * str) {
	gsize len= strlen(str);
	return ShortStrNewLen(str, len);
}

ShortStr * ShortStrNewLen(const gchar * str, gsize len) {
	ShortStr * string;
	gchar * buf;
	Check(len <= 65535);
	string= g_slice_alloc(sizeof(ShortStr) + len);
	string->refcounter= 1;
	string->len= len;
	buf= ShortStrGetStr(string);
	strncpy(buf, str, len);
	buf[len]= 0;
	return string;
}

gchar * ShortStrGetStr(ShortStr * string) {
	return string ? &string->buf[3] : NULL;
}

void ShortStrRef(ShortStr * string) {
	Check(string->refcounter < 255);
	++string->refcounter;
}

void ShortStrUnref(ShortStr * string) {
	--string->refcounter;
	if(string->refcounter==0)
		g_slice_free1(sizeof(ShortStr) + string->len, string);
}

PtrArray * PtrArrayNew(gsize size) {
	PtrArray * array= g_slice_alloc(sizeof(PtrArray) + sizeof(gpointer)*(size - 1));
	array->size= size;
	return array;
}

void PtrArrayFree(PtrArray * array) {
	if(array)
		g_slice_free1(sizeof(PtrArray) + sizeof(gpointer)*(array->size - 1), array);
}

void PtrArrayNull(PtrArray * array) {
	int i;
	for(i= 0; i<array->size; ++i)
		array->ptrs[i]= NULL;
}

GList * CreateListFromPtrArray(PtrArray * array) {
	GList * result= NULL;
	int i;
	for(i= array->size - 1; i >= 0; --i)
		result= g_list_prepend(result, array->ptrs[i]);
	return result;
}

pconstgchar * SplitString(const gchar * str, gchar ch, int * maxcount) {
	int count= 1;
	int length;
	pconstgchar * ptrs;
	int index= 0;
	int i;
	for(length= 0; str[length]; ++length)
		if(str[length]==ch)
			++count;
	if(*maxcount > 0)
		count= MIN(count, *maxcount);
	ptrs= g_slice_alloc(sizeof(gpointer)*(count+1));
	ptrs[index++]= str;
	for(i= 0; i<length && index<count; ++i)
		if(str[i]==ch)
			ptrs[index++]= &str[i+1];
	ptrs[index++]= &str[length+1];
	Check(index==count+1);
	*maxcount= count;
	return ptrs;
}

gsize SplitLength(pconstgchar * ptrs, int i) {
	return ptrs[i+1] - ptrs[i] - 1;
}
 
void FreeSplitPtrs(pconstgchar * ptrs, int count) {
	g_slice_free1(sizeof(gpointer)*(count+1), (gpointer)ptrs);
}

gboolean BoolFromString(const gchar * str) {
	gboolean var= FALSE;
	if(StrEqual(str, "true"))
		var= TRUE;
	else if(StrEqual(str, "false"))
		var= FALSE;
	else
		Check(FALSE);
	return var;
}

const gchar * Translate(const gchar * meta, const gchar * str, gsize * size) {
	const gchar * result= str;

	if(meta && *size > 0) {	

		int count= 3;
		pconstgchar * vstr= SplitString(meta, '|', &count);
		const gchar * prefix= NULL;
		gsize prefix_size= 0;

		if(strncmp(vstr[0], "translate", 9) == 0) {
			if(count > 1) {
				prefix= vstr[1];
				prefix_size= SplitLength(vstr, 1);
			}

			if(prefix_size==0)
				result= gettext(str);
			else {
				gsize buf_size= prefix_size + 1 + *size + 1;
				gchar * buf= g_slice_alloc(buf_size);
				strncpy(buf, prefix, prefix_size); 
				buf[prefix_size]= '|';
				strncpy(&buf[prefix_size+1], str, *size + 1);
				result= gettext(buf);
				g_slice_free1(buf_size, buf);
			}

			*size= strlen(result);
		}

		FreeSplitPtrs(vstr, count);
	}

	return result;
}

gboolean StrEqual(const gchar * s1, const gchar * s2) {
	return strcmp(s1, s2) == 0;
}

void InitInstance(GuiLoader * loader) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	priv->preview= FALSE;
	priv->steal= FALSE;
	priv->stack= NULL;
	priv->model= NULL;
	priv->objects= g_hash_table_new(g_str_hash, g_str_equal);
	priv->connectors= g_hash_table_new(g_str_hash, g_str_equal);
	priv->delayed= NULL;
	priv->strings= NULL;
}

void DoneInstance(GuiLoader * loader) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	FreeGui(loader);
	g_hash_table_destroy(priv->objects);
	g_hash_table_destroy(priv->connectors);
}

void OnStartElement(
	GMarkupParseContext * context,
	const gchar * element,
	const gchar ** attr_names,
	const gchar ** attr_values,
	gpointer user_data,
	GError ** error
) {

	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(user_data);

	Node * node= NewNode();

	if(StrEqual(element, "gui")) {

		int version= 0;
		const gchar * nspace= NULL;

		int i;		
		for(i=0; attr_names[i]; ++i) {
			if(StrEqual(attr_names[i], "namespace"))
				nspace= attr_values[i];
			else if(StrEqual(attr_names[i], "version"))
				version= atoi(attr_values[i]);
			else
				Check(FALSE);
		}

		if(version < 7)
			FatalError("Early GuiXml version detected: resave in Gideon 2.7.0 or later");

		if(version > GUILOADER_VERSION_GUIXML || !StrEqual(nspace, "GTK+"))
			FatalError("Unsupported GuiXml version or namespace");
		
		priv->model= node;
		priv->stack= g_slist_prepend(priv->stack, node);

	} else {
	
		int i;
		for(i=0; attr_names[i]; ++i) {
			if(StrEqual(attr_names[i], "name"))
				node->name= AddSharedStr(user_data, ShortStrNew(attr_values[i]));
			else if(StrEqual(attr_names[i], "type"))
				node->typename= ShortStrNew(attr_values[i]);
			else if(StrEqual(attr_names[i], "meta"))
		    	node->meta= ShortStrNew(attr_values[i]);
			else
				Check(FALSE);
		}

		if(StrEqual(element, "entity"))
			node->role= nrEntity;
		else if(StrEqual(element, "scalar"))
			node->role= nrScalar;
		else if(StrEqual(element, "vector"))
			node->role= nrVector;
		else if(StrEqual(element, "link"))
			node->role= nrLink;
		else
			Check(FALSE);

		node->owner= priv->stack->data;
		g_queue_push_tail(node->owner->domain, node);
		priv->stack= g_slist_prepend(priv->stack, node);

		if(node->owner->role==nrVector) {
			gchar buf[64];
			g_snprintf(buf, sizeof(buf), "%d", g_queue_get_length(node->owner->domain) - 1);
			Check(!node->name);
			node->name= AddSharedStr(user_data, ShortStrNew(buf));
		}

		if(node->role==nrEntity) {
			GuiLoaderInit * init= GUI_LOADER_GET_CLASS(user_data)->init;
			Proxy * proxy= g_hash_table_lookup(init->proxies, ShortStrGetStr(node->typename));
			Check(proxy && !node->proxy);
			node->proxy= proxy;
		}

	}
  
}

void OnEndElement(
	GMarkupParseContext * context,
	const gchar * element,
	gpointer user_data,
	GError ** error
) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(user_data);
	priv->stack= g_slist_delete_link(priv->stack, priv->stack);
}

void OnText(
	GMarkupParseContext * context,
	const gchar * text,
	gsize length,
	gpointer user_data,
	GError ** error
) {
	GuiLoaderPrivate * priv;
	Node * node;
	if(length == 0)
		return;
	priv= GUI_LOADER_GET_PRIVATE(user_data);
	node= priv->stack->data;
	Check(!node->text);
	if(node->role==nrScalar) {
		gsize size= length;
		const gchar * str= priv->preview ? text : Translate(ShortStrGetStr(node->meta), text, &size);
		node->text= ShortStrNewLen(str, size);
	} else if(node->role==nrLink)
		node->text= ShortStrNewLen(text, length);
}

void OnPassthrough(
	GMarkupParseContext * context,
	const gchar * text,
	gsize length,
	gpointer user_data,
	GError ** error
) {
	if(g_str_has_prefix(text, "<![CDATA[\n") && g_str_has_suffix(text, "]]>"))
		OnText(context, text+10, length-10-3, user_data, error);
}

void LoadModel(GuiLoader * loader, const gchar * data, gssize size) {
	GMarkupParseContext * context;
	
	static GMarkupParser parser= {
	  OnStartElement,
	  OnEndElement,
	  OnText,
	  OnPassthrough,
	  NULL
	};

	GError * error= NULL;

	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	Check(!priv->strings);
		
	context= g_markup_parse_context_new(&parser, 0, loader, NULL);
	
	if(!g_markup_parse_context_parse(context, data, size, &error))
		FatalError(error->message);
	
	if(!g_markup_parse_context_end_parse(context, &error))
		FatalError(error->message);

	g_markup_parse_context_free(context);

	Check(!priv->stack);
}

Node * NewNode() {
	Node * node= g_slice_new(Node);
	node->role= nrNone;
	node->name= NULL;
	node->typename= NULL;
	node->meta= NULL;
	node->text= NULL;
	node->owner= NULL;
	node->domain= g_queue_new();
	node->proxy= NULL;
	node->object= NULL;
	return node;
}

void FreeModel(GuiLoader * loader) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	Check(priv->stack==NULL);
	FreeNodeAndDomain(priv->model);
	priv->model= NULL;
}

void FreeNodeAndDomain(Node * node) {
	if(node) {
		GList * elem;
		for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem))
			FreeNodeAndDomain(elem->data);
		FreeNode(node);
	}
}

void FreeNode(Node * node) {
	if(node) {
		if(node->name)
			ShortStrUnref(node->name);
		if(node->typename)
			ShortStrUnref(node->typename);
		if(node->meta)
			ShortStrUnref(node->meta);
		if(node->text)
			ShortStrUnref(node->text);
		g_queue_free(node->domain);
		g_slice_free(Node, node);
	}
}

Node * FindNode(Node * node, const gchar * name) {
	GList * elem;
	for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem)) {
		Node * N= elem->data;
		if(StrEqual(name, ShortStrGetStr(N->name)))
			return N;
	}
	return NULL;
}

ShortStr * AddSharedStr(GuiLoader * loader, ShortStr * string) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	priv->strings= g_slist_prepend(priv->strings, string);
	ShortStrRef(string);
	return string;
}

void LoadGui(GuiLoader * loader) {
	Node * root;
	int updated_count= 0;
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	PropertyFuncData data;
	GSList * elem;
	GList * i1, * i2;

	Check(!priv->delayed);
	root= priv->model;
	g_queue_sort(root->domain, RootOrder, NULL);
	
	i1= g_queue_peek_head_link(root->domain);
	i2= g_queue_peek_head_link(root->domain);
	while(i2) {
		GList * in= g_list_next(i2);
		int p2= ((Node *)i2->data)->proxy->priority;
		int pn= !in ? 100 : ((Node *)in->data)->proxy->priority;
		if(pn > p2) {
			GList * ii;
			for(ii= i1; ; ii= g_list_next(ii)) {
				Node * node= ii->data;
				++updated_count;
				MakeObject(loader, node, NULL);
				g_hash_table_insert(priv->objects, ShortStrGetStr(node->name), node->object);
				if(ii==i2)
					break;
			}
			for(ii= i1; ; ii= g_list_next(ii)) {
				LoadObject(loader, ii->data);
				if(ii==i2)
					break;
			}
			i1= in;
		}
		i2= in;
	}

	Check(updated_count==g_queue_get_length(root->domain));
	
	data.loader= loader;
	for(elem= priv->delayed; elem; elem= g_slist_next(elem)) {
		DelayedAssignment * assignment= elem->data;
		data.object= assignment->object;
		data.name= assignment->property->name;
		data.node= assignment->node;
		assignment->property->assign(&data);
		g_slice_free(DelayedAssignment, assignment);
	}
	g_slist_free(priv->delayed);
	priv->delayed= NULL;

}

gint RootOrder(gconstpointer pa, gconstpointer pb, gpointer usedata) {
	gint a= ((Node *)pa)->proxy->priority;
	gint b= ((Node *)pb)->proxy->priority;
	return a - b;
}

void FreeGui(GuiLoader * loader) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	GSList * elem;
	g_hash_table_foreach_remove(priv->objects, RemoveObject, loader);
	g_hash_table_foreach_remove(priv->connectors, RemoveConnector, loader);
	for(elem= priv->strings; elem; elem= g_slist_next(elem))
		ShortStrUnref(elem->data);
	g_slist_free(priv->strings);
	priv->strings= NULL;
}

gboolean RemoveObject(gpointer key, gpointer value, gpointer userdata) {
	GObject * object= G_OBJECT(value);
	if(GTK_IS_WINDOW(object)) {
		gboolean steal= GUI_LOADER_GET_PRIVATE(userdata)->steal;
		gboolean bound= g_object_get_data(object, "GuiLoaderBoundWindow") != NULL;
		if(!steal || !bound)
			gtk_object_destroy(GTK_OBJECT(object));
	}
	g_object_unref(object);
	return TRUE;
}

gboolean RemoveConnector(gpointer key, gpointer value, gpointer userdata) {
	GQueue * queue= value;
	GList * elem;
	for(elem= g_queue_peek_head_link(queue); elem; elem= g_list_next(elem)) {
		GuiConnector * connector= elem->data;
		ShortStrUnref(connector->emitter->signame);
		ShortStrUnref(connector->emitter->mapname);
		g_slice_free(GideonEmitter, connector->emitter);
		g_slice_free(GuiConnector, connector);
	}
	g_queue_free(queue);
	return TRUE;
}

void MakeObject(GuiLoader * loader, Node * node, GObject * old) {
	GObject * object;
	GList * elem;
	Check(node->proxy && !node->object);

	object= old;
	if(!object) {
		if(node->proxy->createInstance)
			object= node->proxy->createInstance(loader, node);
		else
			object= g_object_new(node->proxy->type, NULL);
	}

	if(g_object_is_floating(object))
		g_object_ref_sink(object);
	else if(GTK_IS_WINDOW(object))
		g_object_ref(object);
		
	for(elem= g_queue_peek_head_link(node->proxy->initInstanceList); elem; elem= g_list_next(elem)) {
		InstanceInitFunc initInstance= elem->data;
		initInstance(object, node);
	}

	node->object= object;
}

void LoadObject(GuiLoader * loader, Node * node) {
	PropertyFuncData data;
	GList * elem;
	Check(node->proxy && node->object);
	data.loader= loader;
	data.object= node->object;
	for(elem= g_queue_peek_head_link(node->proxy->properties); elem; elem= g_list_next(elem)) {
		Property * property= elem->data;
		Node * pnode= FindNode(node, property->name);
		if(pnode) {
			if(!property->lazy) {
				data.name= property->name;
				data.node= pnode;
				property->assign(&data);
			} else
				AddDelayedAssignment(loader, node->object, property, pnode);
		}
	}	
}

void AddConnector(GuiLoader * loader, GObject * object, GideonEmitter * emitter) {
	GuiConnector * connector;
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	GQueue * queue= g_hash_table_lookup(priv->connectors, ShortStrGetStr(emitter->mapname));
	if(!queue) {
		queue= g_queue_new();
		g_hash_table_insert(priv->connectors, ShortStrGetStr(emitter->mapname), queue);
	}
	connector= g_slice_new(GuiConnector);
	connector->object= object;
	connector->emitter= emitter;
	g_queue_push_tail(queue, connector);
}

void AddDelayedAssignment(GuiLoader * loader, GObject * object, Property * property, Node * node) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	DelayedAssignment * assignment= g_slice_new(DelayedAssignment);
	assignment->object= object;
	assignment->property= property;
	assignment->node= node;
	priv->delayed= g_slist_prepend(priv->delayed, assignment);
}

GObject * FindObject(GuiLoader * loader, const gchar * name) {
	GuiLoaderPrivate * priv;
	Check(name);
	priv= GUI_LOADER_GET_PRIVATE(loader);
	return g_hash_table_lookup(priv->objects, name);
}

GuiLoaderInit * NewInit(void) {
	GuiLoaderInit * init= g_slice_new(GuiLoaderInit);
	Check(GTK_MAJOR_VERSION*1000+GTK_MINOR_VERSION <= gtk_major_version*1000+gtk_minor_version);
	init->refcounter= 1;
	init->proxies= g_hash_table_new(g_str_hash, g_str_equal);
	RegisterProxies(init);
	return init;
}

void FreeInit(GuiLoaderInit * init) {
	Check(init->refcounter==0);
	g_hash_table_foreach_remove(init->proxies, RemoveProxy, NULL);
	g_slice_free(GuiLoaderInit, init);
}

gboolean RemoveProxy(gpointer key, gpointer value, gpointer userdata) {
	Proxy * proxy= value;
	GList * elem;
	for(elem= g_queue_peek_head_link(proxy->properties); elem; elem= g_list_next(elem))
		g_slice_free(Property, elem->data);
	g_queue_free(proxy->properties);
	g_queue_free(proxy->initInstanceList);
	g_slice_free(Proxy, proxy);
	return TRUE;
}

void AddProxy(GuiLoaderInit * init, GType type, ProxyCreateFunc create, int priority) {
	Proxy * proxy= g_slice_new(Proxy);
	proxy->type= type;
	proxy->properties= g_queue_new();
	proxy->createInstance= NULL;
	proxy->initInstanceList= g_queue_new();
	proxy->bases= 0;
	proxy->priority= priority;
	create(proxy);
	g_hash_table_insert(init->proxies, (gpointer)g_type_name(type), proxy);
}

void AddProperty(Proxy * proxy, const gchar * name, PropertyFunc func) {
	Property * property;
	#ifdef GUILOADER_SANDBOX 
		Check(!FindProperty(proxy, name));
	#endif
	property= g_slice_new(Property);
	property->name= name;
	property->assign= func;
	property->lazy= FALSE;
	g_queue_push_tail(proxy->properties, property);
}

Property * FindProperty(Proxy * proxy, const gchar * name) {
	GList * elem;
	for(elem= g_queue_peek_head_link(proxy->properties); elem; elem= g_list_next(elem)) {
		Property * property= elem->data;
		if(StrEqual(name, property->name))
			return property;
	}
	return NULL;
}

void SetPropertyFunc(Proxy * proxy, const gchar * name, PropertyFunc func) {
	Property * property= FindProperty(proxy, name);
	property->assign= func;
}

void SetPropertyLazy(Proxy * proxy, const gchar * name, gboolean lazy) {
	Property * property= FindProperty(proxy, name);
	property->lazy= lazy;
}

void AddInitInstance(Proxy * proxy, InstanceInitFunc func) {
	g_queue_push_tail(proxy->initInstanceList, func);
}

void SetCreateInstance(Proxy * proxy, InstanceCreateFunc func) {
	proxy->createInstance= func;
}

void CheckNodeScalar(Node * node) {
	Check(node->role==nrScalar);
	Check(node->typename);
	Check(g_queue_is_empty(node->domain));
}

void CheckNodeVector(Node * node) {
	Check(node->role==nrVector);
	Check(!node->typename);
	Check(!g_queue_is_empty(node->domain));
}

void CheckNodeObject(Node * node) {
	Check(
		(node->role==nrEntity && node->typename) || 
		(node->role==nrLink && !node->typename && g_queue_is_empty(node->domain))
	);
}

gint ModelGetInt(Node * node) {
	gchar * endptr, * str;
	gint var;
	CheckNodeScalar(node);
	str= ShortStrGetStr(node->text);
	var= strtol(str, &endptr, 10);
	Check(*str!=0 && *endptr==0);
	return var;
}

guint ModelGetUnsigned(Node * node) {
	gchar * endptr, * str;
	guint var;
	CheckNodeScalar(node);
	str= ShortStrGetStr(node->text);
	var= strtoul(str, &endptr, 10);
	Check(*str!=0 && *endptr==0);
	return var;
}

gboolean ModelGetBool(Node * node) {
	gboolean var;
	CheckNodeScalar(node);
	var= BoolFromString(ShortStrGetStr(node->text));
	return var;
}

gfloat ModelGetFloat(Node * node) {
	gchar * endptr, * str;
	gfloat var;
	CheckNodeScalar(node);
	str= ShortStrGetStr(node->text);
	var= g_ascii_strtod(str, &endptr);
	Check(*str!=0 && *endptr==0);
	return var;
}

gdouble ModelGetDouble(Node * node) {
	gchar * endptr;
	gchar * str;
	gdouble var;
	CheckNodeScalar(node);
	str= ShortStrGetStr(node->text);
	var= g_ascii_strtod(str, &endptr);
	Check(*str!=0 && *endptr==0);
	return var;
}

gchar * ModelGetString(Node * node) {
	gchar * str;
	CheckNodeScalar(node);
	str= ShortStrGetStr(node->text);
	return str ? str : "";
}

gunichar ModelGetChar(Node * node) {
	gchar * str;
	gunichar var;
	CheckNodeScalar(node);
	str= ShortStrGetStr(node->text);
	Check(g_utf8_strlen(str, -1) == 1);
	var= g_utf8_get_char(str);
	return var;
}

gint ModelGetEnum(Node * node) {
	GType type;
	GEnumClass * meta;
	GEnumValue * evalue;
	gint var;
	CheckNodeScalar(node);
	type= g_type_from_name(ShortStrGetStr(node->typename));
	Check(G_TYPE_IS_ENUM(type));
	meta= G_ENUM_CLASS(g_type_class_ref(type));
	evalue= g_enum_get_value_by_name(meta, ShortStrGetStr(node->text));
	Check(evalue);
	var= evalue->value;
	g_type_class_unref(meta);
	return var;
}

GideonPackOptions ModelGetPackOptions(Node * node) {
	gchar * str;
	GideonPackOptions var;
	CheckNodeScalar(node);
	str= ShortStrGetStr(node->text);
	var= GIDEON_PACK_SHRINK;
	if(StrEqual(str, "GIDEON_PACK_SHRINK"))
		var= GIDEON_PACK_SHRINK;
	else if(StrEqual(str, "GIDEON_PACK_EXPAND_PADDING"))
		var= GIDEON_PACK_EXPAND_PADDING;
	else if(StrEqual(str, "GIDEON_PACK_EXPAND_WIDGET"))
		var= GIDEON_PACK_EXPAND_WIDGET;
	else
		Check(FALSE);
	return var;
}

GideonPoint ModelGetPoint(Node * node) {
	gchar * str;
	GideonPoint point;
	int n=0;
	CheckNodeScalar(node);
	str= ShortStrGetStr(node->text);
	sscanf(str, "%d,%d%n", &point.x, &point.y, &n);
	Check(*str!=0 && *(str + n)==0);
	return point;
}

GdkColor ModelGetColor(Node * node) {
	GdkColor color= { 0, };
	gint result;
	CheckNodeScalar(node);
	result= gdk_color_parse(ShortStrGetStr(node->text), &color);
	Check(result==TRUE);
	return color;
}

GideonEmitter * ModelGetEmitter(Node * node) {
	int count= 3;
	pconstgchar * ptrs;
	GideonEmitter * emitter;
	CheckNodeScalar(node);
	ptrs= SplitString(ShortStrGetStr(node->text), ',', &count);
	Check(count==3);
	emitter= g_slice_new(GideonEmitter);
	emitter->signame= ShortStrNewLen(ptrs[0], SplitLength(ptrs, 0));
	emitter->mapname= ShortStrNewLen(ptrs[1], SplitLength(ptrs, 1));
	emitter->after= BoolFromString(ptrs[2]);
	FreeSplitPtrs(ptrs, count);
	return emitter;
}

GObject * ModelGetObject(GuiLoader * loader, Node * node, GObject * old) {
	CheckNodeObject(node);
	if(node->role==nrLink) {
		const gchar * target= ShortStrGetStr(node->text);
		return target ? FindObject(loader, target) : NULL;
	} else {
		MakeObject(loader, node, old);
		LoadObject(loader, node);
		return node->object;
	}
}

PtrArray * ModelGetObjectVector(GuiLoader * loader, Node * node) {
	PtrArray * result;
	GList * elem;
	int i= 0;
	CheckNodeVector(node);
	result= PtrArrayNew(g_queue_get_length(node->domain));
	for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem)) {
		Node * item= elem->data;
		GObject * object= ModelGetObject(loader, item, NULL);
		result->ptrs[i++]= object;
	}
	return result;
}

void FreeObjectVector(PtrArray * array) {
	int i;
	for(i= 0; i < array->size; ++i) {
		GObject * object= array->ptrs[i];
		if(object)
			g_object_unref(object);
	}
	PtrArrayFree(array);
}

PtrArray * ModelGetStringVector(Node * node) {
	PtrArray * result;
	GList * elem;
	int i= 0;
	CheckNodeVector(node);
	result= PtrArrayNew(g_queue_get_length(node->domain));
	for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem)) {
		Node * item= elem->data;
		gchar * string= ModelGetString(item);
		result->ptrs[i++]= string;
	}
	return result;
}

PtrArray * ModelGetEmitterVector(Node * node) {
	PtrArray * result;
	GList * elem;
	int i= 0;
	CheckNodeVector(node);
	result= PtrArrayNew(g_queue_get_length(node->domain));
	for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem)) {
		Node * item= elem->data;
		GideonEmitter * emitter= ModelGetEmitter(item);
		result->ptrs[i++]=  emitter;
	}
	return result;
}

void PropertyInt(PropertyFuncData * data) {
	gint var= ModelGetInt(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_INT);
	g_value_set_int(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyUnsigned(PropertyFuncData * data) {
	guint var= ModelGetUnsigned(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_UINT);
	g_value_set_uint(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyBool(PropertyFuncData * data) {
	gboolean var= ModelGetBool(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_BOOLEAN);
	g_value_set_boolean(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyFloat(PropertyFuncData * data) {
	gfloat var= ModelGetFloat(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_FLOAT);
	g_value_set_float(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyDouble(PropertyFuncData * data) {
	gdouble var= ModelGetDouble(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_DOUBLE);
	g_value_set_double(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyString(PropertyFuncData * data) {
	const gchar * var= ModelGetString(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_STRING);
	g_value_set_static_string(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyChar(PropertyFuncData * data) {
	gunichar var= ModelGetChar(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_UINT);
	g_value_set_uint(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyEnum(PropertyFuncData * data) {
	gint var= ModelGetEnum(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_INT);
	g_value_set_int(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyObject(PropertyFuncData * data) {
	GObject * old, * var;
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_OBJECT);
	g_object_get_property(data->object, data->name, &value);
	old= g_value_get_object(&value);
	var= ModelGetObject(data->loader, data->node, old);
	Check(old ? old==var : var!=NULL);
	if(var!=old) {
		g_value_set_object(&value, var);
		g_object_set_property(data->object, data->name, &value);
	}
	g_value_unset(&value);
}

void PropertyColor(PropertyFuncData * data) {
	GdkColor var= ModelGetColor(data->node);
	GValue value= { 0, };
	g_value_init(&value, GDK_TYPE_COLOR);
	g_value_set_static_boxed(&value, &var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}
